from __future__ import annotations

from .server_extension import (
    SERVER_APPS_READ_PERMISSION_ID,
    SERVER_EXTENSION_TYPE,
    SERVER_SESSIONS_READ_PERMISSION_ID,
    SERVER_SHUTDOWN_PERMISSION_ID,
    TRUSTED_ORIGINS_GET_PERMISSION_ID,
    TRUSTED_ORIGINS_SET_PERMISSION_ID,
    ServerExtension,
)

__all__ = [
    "SERVER_APPS_READ_PERMISSION_ID",
    "SERVER_EXTENSION_TYPE",
    "SERVER_SESSIONS_READ_PERMISSION_ID",
    "SERVER_SHUTDOWN_PERMISSION_ID",
    "TRUSTED_ORIGINS_GET_PERMISSION_ID",
    "TRUSTED_ORIGINS_SET_PERMISSION_ID",
    "ServerExtension",
]
